# Uses the trained ANN from Expanse. Trained ANN is generated by ANN_optimization_and_training.py.
# Makes predictions for all the test data.

import operator
import os
import pandas as pd
import sklearn.preprocessing
import sklearn.utils
import numpy as np
from sklearn.ensemble import RandomForestRegressor
import sklearn.metrics as sklm
import sklearn.gaussian_process.kernels
import pickle
import json
import sys
from sklearn.metrics import r2_score, mean_absolute_error
from sklearn.kernel_ridge import KernelRidge
from hyperopt import hp, tpe, fmin, Trials
from functools import partial
from molSimplifyAD.retrain.nets import build_ANN, auc_callback, cal_auc, compile_model
from molSimplifyAD.retrain.model_optimization import train_model_hyperopt
import keras
from keras import backend as K
import tensorflow as tf

descriptors = ['CN_IP', 'CN_EA', 'NN_IP', 'NN_EA', 'CN_HOMO', 'CN_LUMO', 'NN_HOMO', 'NN_LUMO', 'lCN_C_mull', 'lCN_N_mull', 'lNN_N1_mull', 'lNN_N2_mull'] # l: ligand only 

# Enforce a floor of zero.
def nonnegative_filter(predictions):
    for count, value in enumerate(predictions):
        if value < 0:
            predictions[count] = 0 # set to zero if prediction is negative
    return predictions

# Data normalization.
def normalize_data(df_train, df_test, fnames, lname, unit_trans=1, debug=False):
    _df_train = df_train.copy().dropna(subset=fnames + lname)
    _df_test = df_test.copy().dropna(subset=fnames + lname)
    X_train, X_test = _df_train[fnames].values, _df_test[fnames].values
    y_train, y_test = _df_train[lname].values * unit_trans, _df_test[lname].values * unit_trans
    if debug:
        print("training data reduced from %d -> %d because of nan." %
              (len(df_train), y_train.shape[0]))
        print("test data reduced from %d -> %d because of nan." %
              (len(df_test), y_test.shape[0]))
    x_scaler = sklearn.preprocessing.StandardScaler()
    x_scaler.fit(X_train)
    X_train = x_scaler.transform(X_train)
    X_test = x_scaler.transform(X_test)
    y_scaler = sklearn.preprocessing.StandardScaler()
    y_scaler.fit(y_train)
    y_train = y_scaler.transform(y_train)
    y_test = y_scaler.transform(y_test)
    for ii, ln in enumerate(lname):
        if debug:
            print(("mean in target: %s, train/test: %f/%f" %
                   (ln, np.mean(y_train[:, ii]), np.mean(y_test[:, ii]))))
    return X_train, X_test, y_train, y_test, x_scaler, y_scaler

# Getting the train, validation, and test data.
df = pd.read_csv("all_data.csv")
df_train = df[df["type"] == "train"]
df_val = df[df["type"] == "val"]
df_train_all = df_train.append(df_val)
df_test = df[df["type"] == "test"]

fnames = descriptors
lname = ['Spectral_Integral'] # The target property that is being predicted for.

# Normalize with train and validation data.
X_train, X_test, y_train, y_test, x_scaler, y_scaler = normalize_data(
    df_train_all, df_test, fnames, lname, unit_trans=1, debug=True)    

# Loading the model generated by ANN_optimization_and_training.py.
model = keras.models.load_model('model_Ir_phosphors.h5', compile=False)

# Applying the ANN to make predictions.
predictions = y_scaler.inverse_transform(model.predict(X_test))
predictions = nonnegative_filter(predictions)

### Will append the predictions to the dataframe df_test, and save to an Excel file.
# Making the predictions into a pandas dataframe, in order to append.
df_predictions = pd.DataFrame(predictions, columns = ['Prediction']) 

df_predictions.reset_index(drop=True, inplace=True)
df_test.reset_index(drop=True, inplace=True)

# Combining test data with the predictions, column wise.
df_predictions = pd.concat([df_predictions, df_test], axis=1) 

# Reordering the columns of the dataframe.
column_order = ['ID', 'Prediction', 'Em50_50', 'Lifetime', 'Sigma', 'Spectral_Integral', 'type'] + descriptors
df_predictions = df_predictions[column_order]

# Writing it all to a CSV.
df_predictions.to_csv("ANN_predictions.csv", index = False)

### Getting the mean absolute error, and the standard deviation of errors
errors = []
for index, row in df_predictions.iterrows():
    errors.append(row['Prediction'] - row['Spectral_Integral']) # change as needed

import statistics
abs_errors = [abs(item) for item in errors]
MAE = statistics.mean(abs_errors)
stdev = statistics.stdev(errors)

print(f'The MAE is {MAE:.4f} and the standard deviation of errors is {stdev:.4f}') # :.4f gives four decimal places
 